"""
.. _crawl_strategies:

🕷 SEO Crawling & Scraping: Strategies & Recipes
================================================

Once you have mastered the basics of using the :ref:`crawl <crawl>` function,
you probably want to achieve more with better customization and control.

These are some code strategies that might be useful to customize how you run
your crawls.

Most of these options can be set using the ``custom_settings`` parameter that
the function takes. This can be set by using a dictionary, where the keys
indicate the option you want to set, and the values specify how you want to set
them.

How to crawl a list of pages, and those pages only (list mode)?
***************************************************************

Simply provide that list as the first argument, for the ``url_list`` parameter,
and make sure that ``follow_links=False``, which is the default. This simply
crawls the given pages, and stops when done.

.. code-block:: python

    >>> import advertools as adv
    >>> url_list = ['https://example.com/page_1',
    ...             'https://example.com/page_2',
    ...             'https://example.com/page_3',
    ...             'https://example.com/page_4']

    >>> adv.crawl(url_list,
    ...           output_file='example_crawl_1.jl',
    ...           follow_links=False)



How can I crawl a website including its sub-domains?
****************************************************

The :ref:`crawl <crawl>` function takes an optional ``allowed_domains``
parameter. If not provided, it defaults to the domains of the URLs in
``url_list``. When the crawler goes through the pages of `example.com`, it
follows links to discover pages. If it finds pages on `help.exmaple.com` it
won't crawl them (it's a different domain). The solution, therefore, is to
provide a list of domains to the ``allowed_domains`` parameter. Make sure you
also include the original domain, in this case `example.com`.

.. code-block:: python

    >>> adv.crawl('https://example.com',
    ...           'example_crawl_1.jl',
    ...           follow_links=True
    ...           allowed_domains=['help.example.com', 'example.com', 'community.example.com'])

How can I save a copy of the logs of my crawl for auditing them later?
**********************************************************************

It's usually good to keep a copy of the logs of all your crawls to check for
errors, exceptions, stats, etc.
Pass a path of the file where you want the logs to be saved, in a dictionary to
the ``cutom_settings`` parameter.
A good practice for consistency is to give the same name to the ``output_file``
and log file (with a different extension) for easier retreival. For example:

| ``output_file``: 'website_name_crawl_1.jl'
| ``LOG_FILE``: 'website_name_crawl_1.log' (.txt can also work)
| ``output_file``: 'website_name_crawl_2.jl'
| ``LOG_FILE``: 'website_name_crawl_2.log'

.. code-block:: python

    >>> adv.crawl('https://example.com', 'example_crawl_1.jl',
    ...           custom_settings={'LOG_FILE': 'example_crawl_1.log'})

How can I automatically stop my crawl based on a certain condition?
*******************************************************************

There are a few conditions that you can use to trigger the crawl to stop, and
they mostly have descriptive names:

    * ``CLOSESPIDER_ERRORCOUNT``: You don't want to wait three hours for a
      crawl to finish, only to discover that you had errors all over the place.
      Set a certain number of errors to trigger the crawler to stop, so you can
      investigate the issue.

    * ``CLOSESPIDER_ITEMCOUNT``: Anything scraped from a page is an "item", h1,
      title , meta_desc, etc. Set the crawler to stop after getting a certain
      number of items if you want that.

    * ``CLOSESPIDER_PAGECOUNT``: Stop the crawler after a certain number of
      pages have been crawled. This is useful as an exploratory technique,
      especially with very large websites. It might be good to crawl a few
      thousand pages, get an idea on its structure, and then run a full crawl
      with those insights in mind.

    * ``CLOSESPIDER_TIMEOUT``: Stop the crawler after a certain number of
      seconds.

.. code-block:: python

    >>> adv.crawl('https://example.com', 'example_crawl_1.jl',
    ...           custom_settings={'CLOSESPIDER_PAGECOUNT': 500})

How can I (dis)obey robots.txt rules?
*************************************

The crawler obeys robots.txt rules by default. Sometimes you might want to
check the results of crawls without doing that. You can set the
``ROBOTSTXT_OBEY`` setting under ``custom_settings``:

.. code-block:: python

    >>> adv.crawl('https://example.com',
    ...           'example_crawl_1.jl',
    ...           custom_settings={'ROBOTSTXT_OBEY': False})


How do I set my User-agent while crawling?
******************************************

Set this parameter under `custom_settings` dictionary under the key
``USER_AGENT``. The default User-agent can be found by running
`adv.spider.user_agent`

.. code-block:: python

    >>> adv.spider.user_agent # to get the current User-agent
    >>> adv.crawl('http://example.com',
    ...           'example_crawl_1.jl',
    ...           custom_settings={'USER_AGENT': 'YOUR_USER_AGENT'})


How can I control the number of concurrent requests while crawling?
*******************************************************************
Some servers are set for high sensitivity to automated and/or concurrent
requests, that you can quickly be blocked/banned. You also want to be polite
and not kill those servers, don't you?

There are several ways to set that under the ``custom_settings`` parameter.
The available keys are the following:

| ``CONCURRENT_ITEMS``: default 100
| ``CONCURRENT_REQUESTS`` : default 16
| ``CONCURRENT_REQUESTS_PER_DOMAIN``: default 8
| ``CONCURRENT_REQUESTS_PER_IP``: default 0

.. code-block:: python

    >>> adv.crawl('https://example.com',
    ...           'example_crawl_1.jl',
    ...           custom_settings={'CONCURRENT_REQUESTS_PER_DOMAIN': 1})

How can I slow down the crawling so I don't hit the websites' servers too hard?
*******************************************************************************

Use the ``DOWNLOAD_DELAY`` setting and set the interval to be waited before
downloading consecutive pages from the same website (in seconds).

.. code-block:: python

    >>> adv.crawl('https://example.com', 'example_crawl_1.jl',
    ...           custom_settings={'DOWNLOAD_DELAY': 3}) # wait 3 seconds between pages

How can I set multiple settings to the same crawl job?
******************************************************
Simply add multiple settings to the ``custom_settings`` parameter.

.. code-block:: python

    >>> adv.crawl('http://example.com',
    ...           'example_crawl_1.jl',
    ...           custom_settings={'CLOSESPIDER_PAGECOUNT': 400,
    ...                            'CONCURRENT_ITEMS': 75,
    ...                            'LOG_FILE': 'output_file.log'})

I want to crawl a list of pages, follow links from those pages, but only to a certain specified depth
*****************************************************************************************************

Set the ``DEPTH_LIMIT`` setting in the ``custom_settings`` parameter. A setting
of 1 would follow links one level after the provided URLs in ``url_list``

    >>> adv.crawl(
    ...     "http://example.com", "example_crawl_1.jl", custom_settings={"DEPTH_LIMIT": 2}
    ... )  # follow links two levels from the initial URLs, then stop

How do I pause/resume crawling, while making sure I don't crawl the same page twice?
************************************************************************************

There are several reasons why you might want to do this:

* You want to mainly crawl the updates to the site (you already crawled the site).
* The site is very big, and can't be crawled quickly.
* You are not in a hurry, and you also don't want to hit the servers hard, so
  you run your crawl across days for example.
* As an emergency measure (connection lost, battery died, etc.) you can start
  where you left off

Handling this is extremely simple, and all you have to do is simply provide a
path to a new folder. Make sure it is new and empty, and make sure to only use
it for the same crawl job reruns. That's all you have to worry about. The
``JOBDIR`` setting handles this.

.. code-block:: python

    >>> adv.crawl('http://example.com',
    ...           'example_crawl_1.jl',
    ...           custom_settings={'JOBDIR': '/Path/to/en/empty/folder'})

The first time you run the above code and then stop it. Stopping can happen by
accident (lost connection, closed computer, etc.), manually (you hit ctrl+C) or
you used a custom setting option to stop the crawl after a certain number of
pages, seconds, etc.

The second time you want to run this, you simply run the exact same command
again. If you check the folder that was created you can see a few files that
manage the process. You don't need to worry about any of it. But make sure that
folder doesn't get changed manually, rerun the same command as many times as
you need, and the crawler should handle de-duplication for you.

How do I use a proxy while crawling?
************************************

This requires the following simple steps:

* Install the 3rd party package `scrapy-rotating-proxies <https://github.com/TeamHG-Memex/scrapy-rotating-proxies>`_. This package handles
  the proxy rotation for you, in addition to retries, so you don't need to
  worry about those details.
* Get a list of proxies and save in a text file, one proxy per line
* Set a few ``custom_settings`` in the crawl function
  (``DOWNLOADER_MIDDLEWARES`` and ``ROTATING_PROXY_LIST_PATH``)

.. code-block:: bash

    $ pip install scrapy-rotating-proxies

Save a list of proxies in a text file with the template:

https://username:password@IPADDRESS:PORT

proxies.txt example file (randome values):

https://user123:password123@12.34.56.78:1111
https://user123:password123@12.34.56.78:1112
https://user123:password123@12.34.56.78:1113
https://user123:password123@12.34.56.78:1114


Then, you need to set a few ``custom_settings`` in the crawl function:


.. code-block:: python

    adv.crawl(
        "https://example.com",
        "output_file.jl",
        follow_links=True,
        custom_settings={
            "DOWNLOADER_MIDDLEWARES": {
                "rotating_proxies.middlewares.RotatingProxyMiddleware": 610,
                "rotating_proxies.middlewares.BanDetectionMiddleware": 620,
            },
            "ROTATING_PROXY_LIST_PATH": "proxies.txt",
        },
    )

You can then read the output file normally and see that the proxies are being
used:


.. code-block:: python

    crawldf = pd.read_json("output_file.jl", lines=True)
    crawldf.filter(regex="proxy").head()


====  ============================  =================  =====================================  ===================
  ..  proxy                           _rotating_proxy  request_headers_proxy-authorization      proxy_retry_times
====  ============================  =================  =====================================  ===================
   0  https://123.456.789.101:8893                  1  Basic b3VzY214dHg6ODlld29rMGRsdfgt                     nan
   1  https://123.456.789.101:8894                  1  Basic b3VzY214dHg6ODlld29rMGRsdfgt                     nan
   2  https://123.456.789.101:8895                  1  Basic b3VzY214dHg6ODlld29rMGRsdfgt                     nan
   3  https://123.456.789.101:8896                  1  Basic b3VzY214dHg6ODlld29rMGRsdfgt                     nan
   4  https://123.456.789.101:8897                  1  Basic b3VzY214dHg6ODlld29rMGRsdfgt                     nan
====  ============================  =================  =====================================  ===================

How can I change the default request headers?
*********************************************

This is a very common use case, and it is very easy to do. Simply add the
``DEFAULT_REQUEST_HEADERS`` setting as a dictionary to the ``custom_settings``
parameter:

.. code-block:: python

    adv.crawl(
        url_list="https://example.com",
        output_file="output.jl",
        custom_settings={
            "DEFAULT_REQUEST_HEADERS": {
                "Accept-Language": "es",
                "Accept-Encoding": "gzip, deflate",
            }
        },
    )

You can easily check for the actual request headers that were used while
crawling. In the crawl DataFrame, simply use the regex pattern
``request_headers_``:

.. code-block:: python

    crawldf = pd.read_json("output.jl", lines=True)
    crawldf.filter(regex="request_headers_")

====  =================================  =================================  ============================
  ..  request_headers_accept-language    request_headers_accept-encoding    request_headers_user-agent
====  =================================  =================================  ============================
   0  es                                 gzip, deflate                      advertools/0.13.2
====  =================================  =================================  ============================

XPath expressions for custom extraction
***************************************

The following are some expressions you might find useful in your crawling,
whether you use ``advertools`` or not. The first column indicates whether or
not the respective expression is used by default by the ``advertools`` crawler.

====================  ===================  =================================================================  ===============================================================================================================
Used by advertools    Suggested Name       XPath Expression                                                   What it does
====================  ===================  =================================================================  ===============================================================================================================
True                  title                //title/text()                                                     Extract the text of title tags
True                  meta_desc            //meta[@name='description']/@content                               Extract the content attribute of the meta tag which has the name 'description'
True                  viewport             //meta[@name='viewport']/@content                                  Extract the content attribute of the meta tag which has the name 'viewport
True                  charset              //meta[@charset]/@charset                                          Get the meta tag that has the attribute 'charset', extract the charset attribute
True                  h1                   //h1/text()                                                        Get the h1 tags, extract their text
True                  h2                   //h2/text()                                                        Get the h2 tags, extract their text
True                  h3                   //h3/text()                                                        Get the h3 tags, extract their text
True                  h4                   //h4/text()                                                        Get the h4 tags, extract their text
True                  h5                   //h5/text()                                                        Get the h5 tags, extract their text
True                  h6                   //h6/text()                                                        Get the h6 tags, extract their text
True                  canonical            //link[@rel='canonical']/@href                                     Get the link elements with the rel attribute 'canonical', extract their href attributes
True                  alt_href             //link[@rel='alternate']/@href                                     Get the link elements with the rel attribute 'alternate', extract their href attributes
True                  alt_hreflang         //link[@rel='alternate']/@hreflang                                 Get the link elements with the rel attribute 'alternate', extract their hreflang attributes
True                  og_props             //meta[starts-with(@property, 'og:')]/@property                    Extract all properties of meta tags whos property attribute starts with 'og:' (OpenGraph)
True                  og_content           //meta[starts-with(@property, 'og:')]/@content                     Extract the content of meta tags whos property attribute starts with 'og:' (OpenGraph)
True                  twtr_names           //meta[starts-with(@name, 'twitter:')]/@name                       Get meta tags who's name starts with 'twitter:' and extract their name attribute
True                  twtr_content         //meta[starts-with(@name, 'twitter:')]/@content                    Get meta tags who's name starts with 'twitter:' and extract their content attribute
False                 iframe_src           //iframe/@src                                                      Get the iframes, and extract their src attribute
False                 gtm_script           //script[contains(@src, 'googletagmanager.com/gtm.js?id=')]/@src   Get the script where the src attribute contains googletagmanager.com/gtm.js?id= and extract its src attribute
False                 gtm_noscript         //iframe[contains(@src, 'googletagmanager.com/ns.html?id=')]/@src  Get the iframes where the src attribute contains googletagmanager.com/ns.html?id= and extract the src attribute
False                 link_rel_rel         //link[@rel]/@rel                                                  Get all the link elements that have a rel attribute, extract the rel attributes
False                 link_rel_href        //link[@rel]/@href                                                 Get all the link elements that have a rel attribute, extract the href attributes
False                 link_rel_stylesheet  //link[@rel='stylesheet']/@href                                    Get all the link elements that have a stylesheet attribute, extract their href attribute
False                 css_links            //link[contains(@href, '.css')]/@href                              Get the link elements where the href attribute contains .css, extract their href attribute
True                  nav_links_text       //nav//a/text()                                                    From the nav element, extract the anchor text of links
True                  nav_links_href       //nav//a/@href                                                     From the nav element, extract all the links
True                  header_links_text    //header//a/text()                                                 From the header, extract the anchor text of links
True                  header_links_href    //header//a/@href                                                  From the header, extract all the links
True                  footer_links_text    //footer//a/text()                                                 From the footer, extract the anchor text of links
True                  footer_links_href    //footer//a/@href                                                  From the footer, extract all the links
False                 js_script_src        //script[@type='text/javascript']/@src                             From script tags where the type is text/javascript, extract the src of the script(s)
False                 js_script_text       //script[@type='text/javascript']/text()                           From script tags where the type is text/javascript, extract the text of the script(s)
False                 script_src           //script//@src                                                     Get the src attribute of any <script> tag
False                 canonical_parent     name(//link[@rel='canonical']/..)                                  Get the name of the parent of the link element that has a rel attribute 'canonical'
====================  ===================  =================================================================  ===============================================================================================================

User-agent strings for use in crawling
**************************************

A simple collection of some of the popular user-agents in case you need to test while
crawling:

========================================================  =========================================================================================================================================================================
Name                                                      User agent string
========================================================  =========================================================================================================================================================================
Amazon 4K Fire TV                                         Mozilla/5.0 (Linux; Android 5.1; AFTS Build/LMY47O) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/41.99900.2250.0242 Safari/537.36
Amazon AFTWMST22                                          Mozilla/5.0 (Linux; Android 9; AFTWMST22 Build/PS7233; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/88.0.4324.152 Mobile Safari/537.36
Amazon Kindle 3                                           Mozilla/5.0 (Linux; U; en-US) AppleWebKit/528.5+ (KHTML, like Gecko, Safari/528.5+) Version/4.0 Kindle/3.0 (screen 600x800; rotate)
Amazon Kindle 4                                           Mozilla/5.0 (X11; U; Linux armv7l like Android; en-us) AppleWebKit/531.2+ (KHTML, like Gecko) Version/5.0 Safari/533.2+ Kindle/3.0+
Amazon Kindle Fire HDX 7                                  Mozilla/5.0 (Linux; Android 4.4.3; KFTHWI Build/KTU84M) AppleWebKit/537.36 (KHTML, like Gecko) Silk/47.1.79 like Chrome/47.0.2526.80 Safari/537.36
Apple TV 4th Gen                                          AppleTV5,3/9.1.1
Apple TV 5th Gen 4K                                       AppleTV6,2/11.1
Apple TV 6th Gen 4K                                       AppleTV11,1/11.1
Apple iPhone 11                                           Mozilla/5.0 (iPhone12,1; U; CPU iPhone OS 13_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/15E148 Safari/602.1
Apple iPhone 12                                           Mozilla/5.0 (iPhone13,2; U; CPU iPhone OS 14_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/15E148 Safari/602.1
Apple iPhone 13 Pro Max                                   Mozilla/5.0 (iPhone14,3; U; CPU iPhone OS 15_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/19A346 Safari/602.1
Apple iPhone 6                                            Mozilla/5.0 (Apple-iPhone7C2/1202.466; U; CPU like Mac OS X; en) AppleWebKit/420+ (KHTML, like Gecko) Version/3.0 Mobile/1A543 Safari/419.3
Apple iPhone 7                                            Mozilla/5.0 (iPhone9,3; U; CPU iPhone OS 10_0_1 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14A403 Safari/602.1
Apple iPhone 7 Plus                                       Mozilla/5.0 (iPhone9,4; U; CPU iPhone OS 10_0_1 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14A403 Safari/602.1
Apple iPhone 8                                            Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1
Apple iPhone 8 Plus                                       Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A5370a Safari/604.1
Apple iPhone SE (3rd generation)                          Mozilla/5.0 (iPhone14,6; U; CPU iPhone OS 15_4 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/19E241 Safari/602.1
Apple iPhone X                                            Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1
Apple iPhone XR (Safari)                                  Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1
Apple iPhone XS (Chrome)                                  Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/69.0.3497.105 Mobile/15E148 Safari/605.1
Apple iPhone XS Max (Firefox)                             Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/13.2b11866 Mobile/16A366 Safari/605.1.15
Bing bot                                                  Mozilla/5.0 (compatible; bingbot/2.0; +http://www.bing.com/bingbot.htm)
Chrome OS-based laptop using Chrome browser (Chromebook)  Mozilla/5.0 (X11; CrOS x86_64 8172.45.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.64 Safari/537.36
Chromecast                                                Mozilla/5.0 (CrKey armv7l 1.5.16041) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.0 Safari/537.36
Google ADT-2                                              Dalvik/2.1.0 (Linux; U; Android 9; ADT-2 Build/PTT5.181126.002)
Google Nexus Player                                       Dalvik/2.1.0 (Linux; U; Android 6.0.1; Nexus Player Build/MMB29T)
Google Pixel                                              Mozilla/5.0 (Linux; Android 7.1.1; Google Pixel Build/NMF26F; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/54.0.2840.85 Mobile Safari/537.36
Google Pixel 2                                            Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2 Build/OPD1.170811.002; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/59.0.3071.125 Mobile Safari/537.36
Google Pixel 3                                            Mozilla/5.0 (Linux; Android 10; Google Pixel 4 Build/QD1A.190821.014.C2; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.108 Mobile Safari/537.36
Google Pixel 4                                            Mozilla/5.0 (Linux; Android 10; Google Pixel 4 Build/QD1A.190821.014.C2; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.108 Mobile Safari/537.36
Google Pixel 5                                            Mozilla/5.0 (Linux; Android 11; Pixel 5 Build/RQ3A.210805.001.A1; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/92.0.4515.159 Mobile Safari/537.36
Google Pixel 6                                            Mozilla/5.0 (Linux; Android 12; Pixel 6 Build/SD1A.210817.023; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/94.0.4606.71 Mobile Safari/537.36
Google Pixel C                                            Mozilla/5.0 (Linux; Android 7.0; Pixel C Build/NRD90M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/52.0.2743.98 Safari/537.36
Google bot                                                Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)
HTC Desire 21 Pro 5G                                      Mozilla/5.0 (Linux; Android 10; HTC Desire 21 pro 5G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.127 Mobile Safari/537.36
HTC One M9                                                Mozilla/5.0 (Linux; Android 6.0; HTC One M9 Build/MRA58K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.98 Mobile Safari/537.3
HTC One X10                                               Mozilla/5.0 (Linux; Android 6.0; HTC One X10 Build/MRA58K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/61.0.3163.98 Mobile Safari/537.36
HTC U20 5G                                                Mozilla/5.0 (Linux; Android 10; Wildfire U20 5G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.136 Mobile Safari/537.36
LG G Pad 7.0                                              Mozilla/5.0 (Linux; Android 5.0.2; LG-V410/V41020c Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/34.0.1847.118 Safari/537.36
Lenovo Yoga Tab 11                                        Mozilla/5.0 (Linux; Android 11; Lenovo YT-J706X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36
Linux-based PC using a Firefox browser                    Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:15.0) Gecko/20100101 Firefox/15.0.1
Mac OS X-based computer using a Safari browser            Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_2) AppleWebKit/601.3.9 (KHTML, like Gecko) Version/9.0.2 Safari/601.3.9
Microsoft Lumia 550                                       Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; RM-1127_16056) AppleWebKit/537.36(KHTML, like Gecko) Chrome/42.0.2311.135 Mobile Safari/537.36 Edge/12.10536
Microsoft Lumia 650                                       Mozilla/5.0 (Windows Phone 10.0; Android 6.0.1; Microsoft; RM-1152) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Mobile Safari/537.36 Edge/15.15254
Microsoft Lumia 950                                       Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; Lumia 950) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/13.1058
Minix NEO X5                                              Mozilla/5.0 (Linux; U; Android 4.2.2; he-il; NEO-X5-116A Build/JDQ39) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30
Nexus 6P                                                  Mozilla/5.0 (Linux; Android 6.0.1; Nexus 6P Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.83 Mobile Safari/537.36
Nintendo 3DS                                              Mozilla/5.0 (Nintendo 3DS; U; ; en) Version/1.7412.EU
Nintendo Switch                                           Mozilla/5.0 (Nintendo Switch; WifiWebAuthApplet) AppleWebKit/601.6 (KHTML, like Gecko) NF/4.0.0.5.10 NintendoBrowser/5.1.0.13343
Nintendo Wii U                                            Mozilla/5.0 (Nintendo WiiU) AppleWebKit/536.30 (KHTML, like Gecko) NX/3.0.4.2.12 NintendoBrowser/4.3.1.11264.US
Nvidia Shield Tablet K1                                   Mozilla/5.0 (Linux; Android 6.0.1; SHIELD Tablet K1 Build/MRA58K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/55.0.2883.91 Safari/537.36
Playstation 4                                             Mozilla/5.0 (PlayStation 4 3.11) AppleWebKit/537.73 (KHTML, like Gecko)
Playstation 5                                             Mozilla/5.0 (PlayStation; PlayStation 5/2.26) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0 Safari/605.1.15
Playstation Vita                                          Mozilla/5.0 (PlayStation Vita 3.61) AppleWebKit/537.73 (KHTML, like Gecko) Silk/3.2
Roku Ultra                                                Roku4640X/DVP-7.70 (297.70E04154A)
Samsung Galaxy S10                                        Mozilla/5.0 (Linux; Android 9; SM-G973U Build/PPR1.180610.011) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Mobile Safari/537.36
Samsung Galaxy S20                                        Mozilla/5.0 (Linux; Android 10; SM-G980F Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.96 Mobile Safari/537.36
Samsung Galaxy S21                                        Mozilla/5.0 (Linux; Android 10; SM-G996U Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Mobile Safari/537.36
Samsung Galaxy S22                                        Mozilla/5.0 (Linux; Android 12; SM-S906N Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119 Mobile Safari/537.36
Samsung Galaxy S6                                         Mozilla/5.0 (Linux; Android 6.0.1; SM-G920V Build/MMB29K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.98 Mobile Safari/537.36
Samsung Galaxy S6 Edge Plus                               Mozilla/5.0 (Linux; Android 5.1.1; SM-G928X Build/LMY47X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.83 Mobile Safari/537.36
Samsung Galaxy S7                                         Mozilla/5.0 (Linux; Android 7.0; SM-G930VC Build/NRD90M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/58.0.3029.83 Mobile Safari/537.36
Samsung Galaxy S7 Edge                                    Mozilla/5.0 (Linux; Android 6.0.1; SM-G935S Build/MMB29K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/55.0.2883.91 Mobile Safari/537.36
Samsung Galaxy S8                                         Mozilla/5.0 (Linux; Android 7.0; SM-G892A Build/NRD90M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/60.0.3112.107 Mobile Safari/537.36
Samsung Galaxy S9                                         Mozilla/5.0 (Linux; Android 8.0.0; SM-G960F Build/R16NW) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.84 Mobile Safari/537.36
Samsung Galaxy Tab A                                      Mozilla/5.0 (Linux; Android 5.0.2; SAMSUNG SM-T550 Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/3.3 Chrome/38.0.2125.102 Safari/537.36
Samsung Galaxy Tab S3                                     Mozilla/5.0 (Linux; Android 7.0; SM-T827R4 Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.116 Safari/537.36
Samsung Galaxy Tab S8 Ultra                               Mozilla/5.0 (Linux; Android 12; SM-X906C Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119 Mobile Safari/537.36
Sony Xperia 1                                             Mozilla/5.0 (Linux; Android 9; J8110 Build/55.0.A.0.552; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/71.0.3578.99 Mobile Safari/537.36
Sony Xperia XZ                                            Mozilla/5.0 (Linux; Android 7.1.1; G8231 Build/41.2.A.0.219; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/59.0.3071.125 Mobile Safari/537.36
Sony Xperia Z4 Tablet                                     Mozilla/5.0 (Linux; Android 6.0.1; SGP771 Build/32.2.A.0.253; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/52.0.2743.98 Safari/537.36
Sony Xperia Z5                                            Mozilla/5.0 (Linux; Android 6.0.1; E6653 Build/32.2.A.0.253) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.98 Mobile Safari/537.36
Windows 10-based PC using Edge browser                    Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.246
Windows 7-based PC using a Chrome browser                 Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.111 Safari/537.36
Xbox One                                                  Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Xbox; Xbox One) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/13.10586
Xbox One S                                                Mozilla/5.0 (Windows NT 10.0; Win64; x64; XBOX_ONE_ED) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.79 Safari/537.36 Edge/14.14393
Xbox Series X                                             Mozilla/5.0 (Windows NT 10.0; Win64; x64; Xbox; Xbox Series X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/48.0.2564.82 Safari/537.36 Edge/20.02
Yahoo! bot                                                Mozilla/5.0 (compatible; Yahoo! Slurp; http://help.yahoo.com/help/us/ysearch/slurp)
Googlebot Smartphone                                      Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/W.X.Y.Z Mobile Safari/537.36 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)
Googlebot Desktop                                         Mozilla/5.0 AppleWebKit/537.36 (KHTML, like Gecko; compatible; Googlebot/2.1; +http://www.google.com/bot.html) Chrome/W.X.Y.Z Safari/537.36
Googlebot-Image                                           Googlebot-Image/1.0
Googlebot-News                                            Googlebot-News
Googlebot-Video                                           Googlebot-Video/1.0
Storebot-Google Desktop                                   Mozilla/5.0 (X11; Linux x86_64; Storebot-Google/1.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/W.X.Y.Z Safari/537.36
Storebot-Google Smartphone                                Mozilla/5.0 (Linux; Android 8.0; Pixel 2 Build/OPD3.170816.012; Storebot-Google/1.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/W.X.Y.Z Mobile Safari/537.36
Google-InspectionTool Mobile                              Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/W.X.Y.Z Mobile Safari/537.36 (compatible; Google-InspectionTool/1.0;)
Google-InspectionTool Desktop                             Mozilla/5.0 (compatible; Google-InspectionTool/1.0;)
GoogleOther                                               Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/W.X.Y.Z Mobile Safari/537.36 (compatible; GoogleOther)
GoogleOther-Image                                         GoogleOther-Image/1.0
GoogleOther-Video                                         GoogleOther-Video/1.0
APIs-Google                                               APIs-Google (+https://developers.google.com/webmasters/APIs-Google.html)
AdsBot-Google-Mobile                                      Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/W.X.Y.Z Mobile Safari/537.36 (compatible; AdsBot-Google-Mobile; +http://www.google.com/mobile/adsbot.html)
AdsBot-Google                                             AdsBot-Google (+http://www.google.com/adsbot.html)
Mediapartners-Google                                      Mediapartners-Google
Google-Safety                                             Google-Safety
FeedFetcher-Google                                        FeedFetcher-Google; (+http://www.google.com/feedfetcher.html)
Google Publisher Center                                   GoogleProducer; (+http://goo.gl/7y4SX)
Google Site Verifier                                      Mozilla/5.0 (compatible; Google-Site-Verification/1.0)
========================================================  =========================================================================================================================================================================

"""  # noqa: E501



"""
Convert a robots.txt file (or list of file URLs) to a table in a CSV format
=============================================================================

**Usage**
---------

.. code-block:: console

    advertools robots [-h] [url ...]

Convert a robots.txt file (or list of file URLs) to a table in a CSV format.

You can provide a web URL, or a local file URL on your local machine, for example:

``file:///Users/path/to/robots.txt``

**Examples**
------------

Single robots.txt file:

.. code-block:: console

    advertools robots https://www.google.com/robots.txt

Multiple robots.txt files:

.. code-block:: console

    advertools robots https://www.google.com/robots.txt https://www.google.jo/robots.txt https://www.google.es/robots.txt

Save to a CSV file using output redirection:

.. code-block:: console

    advertools robots https://www.google.com/robots.txt > google_robots.csv

Run the function for a long list of robots files saved in a text file (robotslist.txt):

.. code-block:: console

    advertools robots < robotslist.txt > multi_robots.csv

**Arguments**
-------------

**Positional arguments:**

- **url**: A robots.txt URL (or a list of URLs). Default: None.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.

Download, parse, and save an XML sitemap to a table in a CSV file
==================================================================

**Usage**
---------

.. code-block:: console

    advertools sitemaps [-h] [-r {0,1}] [-s SEPARATOR] [sitemap_url]

**Positional arguments:**

- **sitemap_url**: The URL of the XML sitemap (regular or sitemap index). Default: None.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-r {0,1}, --recursive {0,1}**: Whether or not to fetch sub-sitemaps if it is a sitemap index file. Default: 1.
- **-s SEPARATOR, --separator SEPARATOR**: The separator with which to separate columns of the output. Default: `,`.

Split a list of URLs into their components: scheme, netloc, path, query, etc.
==============================================================================

**Usage**
---------

.. code-block:: console

    advertools urls [-h] [url_list ...]

**Positional arguments:**

- **url_list**: A list of URLs to parse. Default: None.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.

Crawl a list of known URLs using the HEAD method
=================================================

**Usage**
---------

.. code-block:: console

    advertools headers [-h] [-s [CUSTOM_SETTINGS ...]] [url_list ...] output_file

**Positional arguments:**

- **url_list**: A list of URLs. Default: None.
- **output_file**: Filepath - where to save the output (.jl).

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-s [CUSTOM_SETTINGS ...], --custom-settings [CUSTOM_SETTINGS ...]**: Settings that modify the behavior of the crawler.
  Settings should be separated by spaces, and each setting name and value should be separated by an equal sign `=` without spaces between them.

**Example:**

.. code-block:: console

    advertools headers https://example.com example.jl --custom-settings LOG_FILE=logs.log CLOSESPIDER_TIMEOUT=20

Parse, compress and convert a log file to a DataFrame in the .parquet format
=============================================================================

**Usage**
---------

.. code-block:: console

    advertools logs [-h] [-f [FIELDS ...]] log_file output_file errors_file log_format

**Positional arguments:**

- **log_file**: Filepath - the log file.
- **output_file**: Filepath - where to save the output (.parquet).
- **errors_file**: Filepath - where to save the error lines (.txt).
- **log_format**: The format of the logs. Available defaults are: common, combined, common_with_vhost, nginx_error, apache_error. Supply a special regex instead if you have a different format.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-f [FIELDS ...], --fields [FIELDS ...]**: Provide a list of field names to become column names in the parsed compressed file. Default: None.


Perform a reverse DNS lookup on a list of IP addresses
======================================================

**Usage**
---------

.. code-block:: console

    advertools dns [-h] [ip_list ...]

**Description**
---------------

Perform a reverse DNS lookup on a list of IP addresses.

**Positional arguments:**

- **ip_list**: A list of IP addresses. Default: None.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.

Generate a table of SEM keywords by supplying a list of products and a list of intent words
===========================================================================================

**Usage**
---------

.. code-block:: console

    advertools semkw [-h] [-t [{exact,phrase,modified,broad} ...]] [-l MAX_LEN] [-c {0,1}] [-m {0,1}] [-n CAMPAIGN_NAME] products words

**Description**
---------------

Generate a table of SEM keywords by supplying a list of products and a list of intent words.

**Positional arguments:**

- **products**: A file containing the products that you sell, one per line.
- **words**: A file containing the intent words/phrases that you want to combine with products, one per line.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-t [{exact,phrase,modified,broad} ...]**: Specify match types.
- **-l MAX_LEN, --max-len MAX_LEN**: Number of words to combine with products. Default: 3.
- **-c {0,1}, --capitalize-adgroups {0,1}**: Whether or not to capitalize ad group names in the output file. Default: 1.
- **-m {0,1}, --order-matters {0,1}**: Use combinations and permutations, or just combinations. Default: 1.
- **-n CAMPAIGN_NAME, --campaign-name CAMPAIGN_NAME**: The campaign name.

Get stopwords of the selected language
=======================================

**Usage**
---------

.. code-block:: console

    advertools stopwords [-h] {arabic,azerbaijani,bengali,catalan,chinese,croatian,danish,dutch,english,finnish,french,german,greek,hebrew,hindi,hungarian,indonesian,irish,italian,japanese,kazakh,nepali,norwegian,persian,polish,portuguese,romanian,russian,sinhala,spanish,swedish,tagalog,tamil,tatar,telugu,thai,turkish,ukrainian,urdu,vietnamese}

**Description**
---------------

Get stopwords of the selected language.

**Positional arguments:**

- **{arabic,azerbaijani,bengali,catalan,chinese,croatian,danish,dutch,english,finnish,french,german,greek,hebrew,hindi,hungarian,indonesian,irish,italian,japanese,kazakh,nepali,norwegian,persian,polish,portuguese,romanian,russian,sinhala,spanish,swedish,tagalog,tamil,tatar,telugu,thai,turkish,ukrainian,urdu,vietnamese}**: The language to retrieve stopwords for.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.

Get word counts of a text list optionally weighted by a number list
====================================================================

**Usage**
---------

.. code-block:: console

    advertools wordfreq [-h] [-n NUMBER_LIST] [-r REGEX] [-l PHRASE_LEN] [-s [STOPWORDS ...]] [text_list ...]

**Description**
---------------

Get word counts of a text list optionally weighted by a number list.

**Positional arguments:**

- **text_list**: A text list, one document (sentence, tweet, etc.) per line. Default: None.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-n NUMBER_LIST, --number-list NUMBER_LIST**: Filepath for a file containing the number list, one number per line. Default: None.
- **-r REGEX, --regex REGEX**: A regex to tokenize words. Default: None.
- **-l PHRASE_LEN, --phrase-len PHRASE_LEN**: The phrase (token) length to split words (the `n` in n-grams). Default: 1.
- **-s [STOPWORDS ...], --stopwords [STOPWORDS ...]**: A list of stopwords to exclude. Default: None.

Search for emoji using a regex
===============================

**Usage**
---------

.. code-block:: console

    advertools emoji [-h] regex

**Description**
---------------

Search for emoji using a regex.

**Positional arguments:**

- **regex**: Pattern to search for emoji.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.

Extract structured entities from a text list; emoji, hashtags, mentions
========================================================================

**Usage**
---------

.. code-block:: console

    advertools extract [-h] {emoji,hashtags,mentions} text_list

**Description**
---------------

Extract structured entities from a text list; emoji, hashtags, mentions.

**Positional arguments:**

- **{emoji,hashtags,mentions}**: Entity type to extract.
- **text_list**: Filepath containing the text list, one phrase per line.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.

Tokenize documents (phrases, keywords, tweets, etc.) into tokens of the desired length
======================================================================================

**Usage**
---------

.. code-block:: console

    advertools tokenize [-h] [-l LENGTH] [-s SEPARATOR] [text_list ...]

**Description**
---------------

Tokenize documents (phrases, keywords, tweets, etc.) into tokens of the desired length.

**Positional arguments:**

- **text_list**: Filepath containing the text list, one document (sentence, tweet, etc.) per line. Default: None.

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-l LENGTH, --length LENGTH**: Length of tokens (the `n` in n-grams). Default: 1.
- **-s SEPARATOR, --separator SEPARATOR**: Character to separate the tokens. Default: `,`.

SEO crawler
============

**Usage**
---------

.. code-block:: console

    advertools crawl [-h] [-l FOLLOW_LINKS] [-d [ALLOWED_DOMAINS ...]]
                     [--exclude-url-params [EXCLUDE_URL_PARAMS ...]]
                     [--include-url-params [INCLUDE_URL_PARAMS ...]]
                     [--exclude-url-regex EXCLUDE_URL_REGEX]
                     [--include-url-regex INCLUDE_URL_REGEX]
                     [--css-selectors [CSS_SELECTORS ...]]
                     [--xpath-selectors [XPATH_SELECTORS ...]]
                     [--custom-settings [CUSTOM_SETTINGS ...]]
                     [url_list ...] output_file

**Description**
---------------

SEO crawler to extract and collect information from web pages.

**Positional arguments:**

- **url_list**: One or more URLs to crawl. Default: None.
- **output_file**: Filepath where to save the output (.jl).

**Optional arguments:**

- **-h, --help**: Show this help message and exit.
- **-l FOLLOW_LINKS, --follow-links FOLLOW_LINKS**: Whether or not to follow links encountered on crawled pages. Default: 0.
- **-d [ALLOWED_DOMAINS ...], --allowed-domains [ALLOWED_DOMAINS ...]**: Only follow links on these domains while crawling. Default: None.
- **--exclude-url-params [EXCLUDE_URL_PARAMS ...]**: List of URL parameters to exclude while following links. If a link contains any of these parameters, it will not be followed. Default: None.
- **--include-url-params [INCLUDE_URL_PARAMS ...]**: List of URL parameters to include while following links. If a link contains any of these parameters, it will be followed. Default: None.
- **--exclude-url-regex EXCLUDE_URL_REGEX**: A regular expression pattern to exclude certain URLs from being followed. Default: None.
- **--include-url-regex INCLUDE_URL_REGEX**: A regular expression pattern to include certain URLs to be followed. Default: None.
- **--css-selectors [CSS_SELECTORS ...]**: A dictionary mapping column names to CSS selectors. The selectors will be used to extract the required data/content. Default: None.
- **--xpath-selectors [XPATH_SELECTORS ...]**: A dictionary mapping column names to XPath selectors. The selectors will be used to extract the required data/content. Default: None.
- **--custom-settings [CUSTOM_SETTINGS ...]**: Custom settings to modify the behavior of the crawler. Over 170 settings are available for configuration. See Scrapy documentation for details: https://docs.scrapy.org/en/latest/topics/settings.html. Default: None.

**Examples**
------------

Crawl a website starting from its homepage:

.. code-block:: console

    advertools crawl https://example.com example_output.jl --follow-links 1

Crawl a list of pages in list mode:

.. code-block:: console

    advertools crawl url_1 url_2 url_3 example_output.jl

If you have a long list of URLs in a file (url_list.txt):

.. code-block:: console

    advertools crawl < url_list.txt example_output.jl

Stop crawling after processing 1,000 pages:

.. code-block:: console

    advertools crawl https://example.com example_output.jl --follow-links 1 --custom-settings CLOSESPIDER_PAGECOUNT=1000

"""



"""
Introduction to partitioning text
===================================

The ``partition`` function in ``advertools`` provides a powerful way to split a string
based on a regular expression. Unlike typical string splitting methods that only return the text *between* delimiters, ``partition`` includes the delimiters themselves in the result list. This is particularly useful for tasks where the delimiters are as important as the content they separate.

Core Functionality
------------------

The function takes a ``text`` string, a ``regex`` pattern, and optional ``flags`` from the ``re`` module. It returns a list of strings, alternating between the substrings and the matches.

**Key Features:**

*   **Includes Delimiters:** The matched delimiters are part of the output list.
*   **Regex Powered:** Leverages the full power of regular expressions for defining separators.
*   **Handles Edge Cases:** Correctly processes matches at the beginning or end of the string, and consecutive matches, by including empty strings to represent zero-length parts.

Examples
--------

Let's explore some practical examples:

**1. Basic Splitting by Numbers:**

.. code-block:: python

   import advertools as adv

   text = "abc123def456ghi"
   regex = r"\\d+"
   result = adv.partition(text, regex)
   print(result)
   # Output: ['abc', '123', 'def', '456', 'ghi']

**2. No Match Found:**

If the regex pattern doesn't find any matches, the original string is returned as a single-element list.

.. code-block:: python

   import advertools as adv

   text = "test"
   regex = r"X"
   result = adv.partition(text, regex)
   print(result)
   # Output: ['test']

**3. Handling Consecutive Delimiters and Edge Matches:**

This example shows how ``partition`` handles cases where delimiters are at the start/end or appear consecutively.

.. code-block:: python

   import advertools as adv

   text = ",a,,b,"
   regex = r","
   result = adv.partition(text, regex)
   print(result)
   # Output: ['', ',', 'a', ',', '', ',', 'b', ',', '']

**4. Case-Insensitive Partitioning:**

You can use regex flags, like ``re.IGNORECASE``, for more flexible matching.

.. code-block:: python

   import advertools as adv
   import re

   text = "TestData"
   regex = r"t"
   result = adv.partition(text, regex, flags=re.IGNORECASE)
   print(result)
   # Output: ['', 'T', 'es', 't', 'Data']

Connecting to Other Use Cases
-----------------------------

While ``partition`` is a general-purpose string manipulation tool, its ability to retain delimiters makes it valuable in various contexts. For instance, if you were working with a function that processes Markdown documents (let's imagine a hypothetical ``generate_markdown_chunks`` function), ``partition`` could be used to split a Markdown document by specific structural elements (e.g., headings, code blocks, lists).

Imagine you want to break down a Markdown document into chunks based on heading levels (e.g., ``## ``, ``### ``). The ``partition`` function could be used to identify these headings and the content between them.

.. code-block:: python

   import advertools as adv
   import re

   markdown_text = '''
   # Document Title

   Some introductory text.

   ## Section 1

   Content for section 1.

   ### Subsection 1.1

   Details for subsection 1.1.

   ## Section 2

   Content for section 2.
   '''
   # Regex to match markdown headings (##, ###, etc.)
   # Matches lines starting with one or more '#' followed by a space
   heading_regex = r"^#+\\s"


   # Partition the markdown text by headings
   # Note: This is a simplified example. A robust markdown parser would be more complex.
   chunks = adv.partition(markdown_text, heading_regex, flags=re.MULTILINE)

   # The 'chunks' list would contain alternating text blocks and the matched headings,
   # allowing further processing of each part of the document.
   for i, chunk in enumerate(chunks):
       if re.match(heading_regex, chunk):
           print(f"Heading: {chunk.strip()}")
       else:
           print(f"Content Block {i // 2 + 1}:\\n{chunk.strip()}\\n")

This demonstrates how ``partition`` can be a foundational tool for more complex text processing tasks, such as breaking down structured documents into manageable pieces.

"""

import re


def partition(text, regex, flags=0):
    """Partition a string based on a regex pattern.

    Splits the `text` by all occurrences of `regex`. The resulting list
    includes both the substrings between the matches and the matches
    themselves.

    Parameters
    ----------
    text : str
        The input string to partition.
    regex : str
        The regular expression string to find delimiter strings.
        This pattern defines the separators.
    flags : int, optional
        Regex flags from the `re` module (e.g., `re.IGNORECASE`).
        Defaults to 0 (no flags).

    Returns
    -------
    list[str] : A list of strings, where elements are alternating parts of the original
                string and the substrings matched by the regex.
    - If `regex` does not match, the list contains `text` as its
      only element.
    - If matches occur at the beginning/end of `text`, or are
      consecutive, empty strings may be included to represent zero-length parts.

    Examples
    --------
    >>> partition("abc123def456ghi", r"\\d+")
    ['abc', '123', 'def', '456', 'ghi']
    >>> partition("test", r"X")  # No match
    ['test']
    >>> partition(",a,,b,", r",")
    ['', ',', 'a', ',', '', ',', 'b', ',', '']
    >>> partition("startmiddleend", r"middle")
    ['start', 'middle', 'end']
    >>> partition("delimtextdelim", r"delim")
    ['', 'delim', 'text', 'delim', '']
    >>> partition("TestData", r"t", flags=re.IGNORECASE)
    ['', 'T', 'es', 't', 'Data']
    """
    if text == "":
        return [""]

    capturing_regex = f"({regex})"

    parts = re.split(capturing_regex, text, flags=flags)

    return [part.strip() for part in parts if part.strip()]



"""

Image Crawler and Downloader
============================

**Experimental feature - expect changes**

This is a crawler that downloads all images on a given list of URLs. Using
:func:`crawl_images` is straightforward:

>>> import advertools as adv
>>> adv.crawl_images([URL_1, URL_2, URL_3, ...], "output_dir")

This would go to the supplied URLs and download all images found on those URLs, and
place them in ``output_dir``.

You can set a few conditions to modify the behavior:

* ``min_width``: The minimum width in pixels for an image to be downloaded. This is
  mainly to avoid downloading logos, tracking pixels, navigational elemenst as images,
  and so on.
* ``min_height``: The minimum height in pixels for an image to be downloaded
* ``include_img_regex``: A regular expression that the image path needs to match in
  order for it to be downloaded. In some cases, after checking the patterns of images
  for example, you might want to only download images that contain "sports", or any
  other pattern. Or maybe images of interest are under the /economy/ folder and you only
  want those images.
* ``custom_settings``: Just like other crawl functions, you can set any custom settings
  you want to control the crawler's behavior. Some examples include changing the
  User-agent, (dis)obeying robots.txt rules, and so on. More options and code details
  can be found in the :ref:`crawling strategies <crawl_strategies>` page.

To run the :func:`crawl_images` function you need to set an ``output_dir``. This is
where all images will be downloaded. You also get a summary file with details about the
downloaded images. You can read this file through the special function
:func:`summarize_crawled_imgs` to get a few more details about those images.

>>> adv.summarize_crawled_imgs("path/to/output_dir")

====  ==============================================================================================  ==========================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================
  ..  image_location                                                                                  image_urls
====  ==============================================================================================  ==========================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================
   0  https://www.buzzfeed.com/hannahdobro/dirty-little-industry-secrets?origin=tuh                   https://img.buzzfeed.com/buzzfeed-static/static/user_images/6r1oxXOpC_large.jpg?downsize=120:&output-format=jpg&output-quality=auto
   0  https://www.buzzfeed.com/hannahdobro/dirty-little-industry-secrets?origin=tuh                   https://img.buzzfeed.com/buzzfeed-static/static/2024-03/18/16/asset/fce856744ed8/sub-buzz-1303-1710779249-1.jpg
   0  https://www.buzzfeed.com/hannahdobro/dirty-little-industry-secrets?origin=tuh                   data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7
   0  https://www.buzzfeed.com/hannahdobro/dirty-little-industry-secrets?origin=tuh                   https://img.buzzfeed.com/buzzfeed-static/static/2024-03/18/16/asset/245ecfa321e9/sub-buzz-894-1710779358-1.jpg
   1  https://www.buzzfeed.com/chelseastewart/josh-peck-statement-drake-bell-abuse-claims?origin=tuh  https://img.buzzfeed.com/buzzfeed-static/static/2017-12/12/13/user_images/buzzfeed-prod-web-03/chelseastewart-v2-5590-1513102854-0_large.jpg?downsize=120:&output-format=jpg&output-quality=auto
   1  https://www.buzzfeed.com/chelseastewart/josh-peck-statement-drake-bell-abuse-claims?origin=tuh  https://img.buzzfeed.com/buzzfeed-static/static/2024-03/21/19/asset/ea6298160040/sub-buzz-1093-1711048323-1.jpg?downsize=700%3A%2A&output-quality=auto&output-format=auto
   1  https://www.buzzfeed.com/chelseastewart/josh-peck-statement-drake-bell-abuse-claims?origin=tuh  data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7
   1  https://www.buzzfeed.com/chelseastewart/josh-peck-statement-drake-bell-abuse-claims?origin=tuh  data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAFQAAAA7CAMAAADSF118AAAAP1BMVEUAAADIGxPOHBK5EwDFGhi5Fwi8GRTEGhe7EQDMHR7////vyMfddnm5Hx334+Py8fHdj5DLVVXnq6zJOTzVbG1s8SkwAAAACXRSTlMAv4Eo10JnqA8IHfydAAABJUlEQVRYw93Y64rCMBCG4czk5FSzdav3f63bDaxfV4Qm+AXR96/wMNj0kLhtPib9LcutYA8K+F1rKXqH4KmIPZVIOvwnszEqumFjMVLB3+YsRiv8zRqMWHa1ZNQiBuUV3Jo3cn5FlY3qimY2KitajB3+UmLRxRGovgmqTj4HXc69aN5Hj9PcyYqzfXSavk58tJMNTWgv24pW9kpE0fGbioKlomCZKNgLEUXLhYiiMx+dT+xJ8SxgoCDZ6EJcp7jsPBQLlIbiVmpEwy7aS1poeZ30PvqlAQVJRGeQtLfp1dBLPyb0bdDER+OYL2nHR7E34yUjtjw6ZMc3am/KXlSpoodCHrQWiWbxI85Q6Kc9pneHSCmHJ0VJGPPuAC3LWqO/OURL0aEfg76m8Izrt6EAAAAASUVORK5CYII=
   2  https://www.buzzfeed.com/josephlongo/celebs-wearing-rewearing-same-dress?origin=tuh             https://img.buzzfeed.com/buzzfeed-static/static/2021-06/3/16/user_images/a824550933a9/tomiobaro-v2-2174-1622738336-41_large.jpg?downsize=120:&output-format=jpg&output-quality=auto
   2  https://www.buzzfeed.com/josephlongo/celebs-wearing-rewearing-same-dress?origin=tuh             https://img.buzzfeed.com/buzzfeed-static/static/2024-03/19/13/asset/6634db63f453/sub-buzz-576-1710855734-6.jpg?downsize=700%3A%2A&output-quality=auto&output-format=auto
   2  https://www.buzzfeed.com/josephlongo/celebs-wearing-rewearing-same-dress?origin=tuh             https://img.buzzfeed.com/buzzfeed-static/static/2024-03/19/13/asset/cb8db05df7e7/sub-buzz-1743-1710855790-4.jpg
   2  https://www.buzzfeed.com/josephlongo/celebs-wearing-rewearing-same-dress?origin=tuh             data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7
====  ==============================================================================================  ==========================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================================

Image file names
----------------

The downloaded images need to be given a name naturally, and the name is taken from the
slug of the image URL, excluding any query parameters or slashes.
The full URLs of those images can be found in the summary file, and you can access them
through :func:`summarize_crawled_imgs`. You also see where those images are located as
you can see in the table above.

"""  # noqa: E501

import json
import re
import subprocess
from urllib.parse import urlsplit

import pandas as pd
from scrapy import Field, Item, Request, Spider
from scrapy.pipelines.images import ImagesPipeline

import advertools as adv

image_spider_path = adv.__path__[0] + "/image_spider.py"

user_agent = f"advertools/{adv.__version__}"


class ImgItem(Item):
    image_urls = Field()
    images = Field()
    image_location = Field()


class AdvImagesPipeline(ImagesPipeline):
    def file_path(self, request, response=None, info=None, *, item=None):
        img_url = request.url
        return urlsplit(img_url).path.split("/")[-1]


class ImageSpider(Spider):
    name = "image_spider"
    include_img_regex = None
    custom_settings = {
        "USER_AGENT": user_agent,
        "ROBOTSTXT_OBEY": True,
        "HTTPERROR_ALLOW_ALL": True,
        "ITEM_PIPELINES": {AdvImagesPipeline: 1},
        "AUTOTHROTTLE_ENABLED": True,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 8,
    }

    def __init__(self, start_urls, include_img_regex=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_urls = json.loads(json.dumps(start_urls.split(",")))
        if include_img_regex is not None:
            self.include_img_regex = include_img_regex

    def start_requests(self):
        for url in self.start_urls:
            yield Request(url, callback=self.parse)

    def parse(self, response):
        img_item = ImgItem()
        img_src = response.xpath("//img/@src").getall()
        if self.include_img_regex is not None:
            img_src = [
                response.urljoin(src)
                for src in img_src
                if re.findall(self.include_img_regex, src)
            ]
        else:
            img_src = [response.urljoin(src) for src in img_src]
        img_item["image_urls"] = img_src
        img_item["image_location"] = response.request.url
        yield img_item


def crawl_images(
    start_urls,
    output_dir,
    min_width=0,
    min_height=0,
    include_img_regex=None,
    custom_settings=None,
):
    """Download all images available on start_urls and save them to output_dir.

    THIS FUNCTION IS STILL EXPERIMENTAL. Expect many changes.

    Parameters
    ----------

    start_urls : list
      A list of URLs from which you want to download available images.
    output_dir : str
      The directory where you want the images to be saved.
    min_width : int
      The minimum width in pixels for an image to be downloaded.
    min_height : int
      The minimum height in pixels for an image to be downloaded.
    include_img_regex : str
      A regular expression to select image src URLs. Use this to restrict image
      files that match this regex.
    custom_settings : dict
      Additional settings to customize the crawling behaviour.
    """
    settings_list = []
    if custom_settings is not None:
        for key, val in custom_settings.items():
            if isinstance(val, dict):
                setting = "=".join([key, json.dumps(val)])
            else:
                setting = "=".join([key, str(val)])
            settings_list.extend(["-s", setting])

    command = [
        "scrapy",
        "runspider",
        image_spider_path,
        "-a",
        "start_urls=" + ",".join(start_urls),
        "-s",
        "IMAGES_STORE=" + output_dir,
        "-s",
        "IMAGES_MIN_HEIGHT=" + str(min_height),
        "-s",
        "IMAGES_MIN_WIDTH=" + str(min_width),
        "-o",
        output_dir + "/image_summary.jl",
    ] + settings_list
    if include_img_regex is not None:
        command += ["-a", "include_img_regex=" + include_img_regex]
    subprocess.run(command)


def summarize_crawled_imgs(image_dir):
    """Provide a DataFrame of image locations and image URLs resulting from
    crawl_images.

    Running the crawl_images function create a summary CSV file of the
    downloaded images. This function parses that file and provides a two-column
    DataFrame:

    - image_location: The URL from which the images was downloaded from.
    - image_urls: The URL of the image file tha was downloaded.

    Parameters
    ----------
    image_dir : str
      The path to the directory that you provided to crawl_images
    """
    df = pd.read_json(image_dir.rstrip("/") + "/image_summary.jl", lines=True)
    return df[["image_location", "image_urls"]].explode("image_urls")



import re
from unicodedata import name

APOSTROPHE = []
BRACKET = []
COLON = []
COMMA = []
CURRENCY = []
CURRENCY_RAW = []
EXCLAMATION_MARK_RAW = []
FULL_STOP = []
HASHTAG = []
HASHTAG_RAW = []
MENTION = []
MENTION_RAW = []
PAREN = []
QUESTION_MARK_RAW = []
QUOTE = []
SENTENCE_END = []
URL = []
URL_RAW = []
WORD_DELIM = []

for i in range(2_000_000):
    try:
        if "APOSTROPHE" in name(chr(i)) and (chr(i) not in ["ŉ", "\U000e0027"]):
            APOSTROPHE.append(chr(i))
        if (
            "BRACKET" in name(chr(i))
            and "IDEOGRAPH" not in name(chr(i))
            and "TORTOISE SHELL BRACKETED LATIN CAPITAL LETTER S" not in name(chr(i))
        ):
            BRACKET.append(chr(i))
        if "COLON" in name(chr(i)) and i != 8353:  # remove the colon currency sign (₡)
            COLON.append(chr(i))
        if (
            ("COMMA" in name(chr(i)))
            and not re.match("LATIN (SMALL|CAPITAL) LETTER", name(chr(i)))
            and not re.match("DIGIT", name(chr(i)))
        ):
            COMMA.append(chr(i))
        if "EXCLAMATION" in name(chr(i)):
            EXCLAMATION_MARK_RAW.append(chr(i))
        if (
            "FULL STOP" in name(chr(i))
            and (not name(chr(i)).startswith("DIGIT"))
            and (not name(chr(i)).startswith("NUMBER"))
        ):
            FULL_STOP.append(chr(i))
        if "QUOT" in name(chr(i)) and name(chr(i)) != "YI SYLLABLE QUOT":
            QUOTE.append(chr(i))
        if "CURRENC" in name(chr(i)):
            CURRENCY.append(chr(i))
        if ("PAREN" in name(chr(i))) and not re.match("PARENTHESIZED", name(chr(i))):
            PAREN.append(chr(i))
        if "QUESTION" in name(chr(i)) and "IDEOGRAPH" not in name(chr(i)):
            QUESTION_MARK_RAW.append(chr(i))
    except Exception:
        continue
